package anywheresoftware.b4a.agraham.encryption;

import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.ShortName;

/**
 * The Base64 object encodes and decodes to and from Base64 notation.
 */
@ShortName("Base64")
@Author("Andrew Graham")
public class Base64
{
	// 1.0 initial release
	private final static double version = 1.0;
	
   
	// Private fields    
    
     /** Insert break lines when encoding. */
    private boolean breaklines = false;
	
    /** 
     * Encode using Base64-like encoding that is URL- and Filename-safe as described
     * in Section 4 of RFC3548: 
     * <a href="http://www.faqs.org/rfcs/rfc3548.html">http://www.faqs.org/rfcs/rfc3548.html</a>.
     * It is important to note that data encoded this way is <em>not</em> officially valid Base64, 
     * or at the very least should not be called Base64 without also specifying that is
     * was encoded using the URL- and Filename-safe dialect.
     */
    private boolean urlsafe = false;
    
    /** Maximum line length ) of Base64 output. */
    private int maxlinelength = 76;
    
    
    /** The equals sign (=) as a byte. */
    private final static byte EQUALS_SIGN = (byte)'=';
    
    
    /** The new line character (\n) as a byte. */
    private final static byte NEW_LINE = (byte)'\n';
    
 	
    private final static byte WHITE_SPACE_ENC = -5; // Indicates white space in encoding
    private final static byte EQUALS_SIGN_ENC = -1; // Indicates equals sign in encoding
	
	
    // Standard Base64 alphabet
    
    private final static byte[] _STANDARD_ALPHABET = {
        (byte)'A', (byte)'B', (byte)'C', (byte)'D', (byte)'E', (byte)'F', (byte)'G',
        (byte)'H', (byte)'I', (byte)'J', (byte)'K', (byte)'L', (byte)'M', (byte)'N',
        (byte)'O', (byte)'P', (byte)'Q', (byte)'R', (byte)'S', (byte)'T', (byte)'U', 
        (byte)'V', (byte)'W', (byte)'X', (byte)'Y', (byte)'Z',
        (byte)'a', (byte)'b', (byte)'c', (byte)'d', (byte)'e', (byte)'f', (byte)'g',
        (byte)'h', (byte)'i', (byte)'j', (byte)'k', (byte)'l', (byte)'m', (byte)'n',
        (byte)'o', (byte)'p', (byte)'q', (byte)'r', (byte)'s', (byte)'t', (byte)'u', 
        (byte)'v', (byte)'w', (byte)'x', (byte)'y', (byte)'z',
        (byte)'0', (byte)'1', (byte)'2', (byte)'3', (byte)'4', (byte)'5', 
        (byte)'6', (byte)'7', (byte)'8', (byte)'9', (byte)'+', (byte)'/'
    };
	
    
    // Translates a Base64 value to either its 6-bit reconstruction value
    // or a negative number indicating some other meaning.
    private final static byte[] _STANDARD_DECODABET = {
        -9,-9,-9,-9,-9,-9,-9,-9,-9,                 // Decimal  0 -  8
        -5,-5,                                      // Whitespace: Tab and Linefeed
        -9,-9,                                      // Decimal 11 - 12
        -5,                                         // Whitespace: Carriage Return
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 14 - 26
        -9,-9,-9,-9,-9,                             // Decimal 27 - 31
        -5,                                         // Whitespace: Space
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,              // Decimal 33 - 42
        62,                                         // Plus sign at decimal 43
        -9,-9,-9,                                   // Decimal 44 - 46
        63,                                         // Slash at decimal 47
        52,53,54,55,56,57,58,59,60,61,              // Numbers zero through nine
        -9,-9,-9,                                   // Decimal 58 - 60
        -1,                                         // Equals sign at decimal 61
        -9,-9,-9,                                      // Decimal 62 - 64
        0,1,2,3,4,5,6,7,8,9,10,11,12,13,            // Letters 'A' through 'N'
        14,15,16,17,18,19,20,21,22,23,24,25,        // Letters 'O' through 'Z'
        -9,-9,-9,-9,-9,-9,                          // Decimal 91 - 96
        26,27,28,29,30,31,32,33,34,35,36,37,38,     // Letters 'a' through 'm'
        39,40,41,42,43,44,45,46,47,48,49,50,51,     // Letters 'n' through 'z'
        -9,-9,-9,-9,-9                              // Decimal 123 - 127
        ,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,       // Decimal 128 - 139
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 140 - 152
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 153 - 165
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 166 - 178
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 179 - 191
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 192 - 204
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 205 - 217
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 218 - 230
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 231 - 243
        -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9         // Decimal 244 - 255 
    };
	
	
    // URL and Filename safe Base64 alphabet
	
    /**
     * Used in the URL- and Filename-safe dialect described in Section 4 of RFC3548: 
     * <a href="http://www.faqs.org/rfcs/rfc3548.html">http://www.faqs.org/rfcs/rfc3548.html</a>.
     * Notice that the last two bytes become "hyphen" and "underscore" instead of "plus" and "slash."
     */
    private final static byte[] _URL_SAFE_ALPHABET = {
      (byte)'A', (byte)'B', (byte)'C', (byte)'D', (byte)'E', (byte)'F', (byte)'G',
      (byte)'H', (byte)'I', (byte)'J', (byte)'K', (byte)'L', (byte)'M', (byte)'N',
      (byte)'O', (byte)'P', (byte)'Q', (byte)'R', (byte)'S', (byte)'T', (byte)'U', 
      (byte)'V', (byte)'W', (byte)'X', (byte)'Y', (byte)'Z',
      (byte)'a', (byte)'b', (byte)'c', (byte)'d', (byte)'e', (byte)'f', (byte)'g',
      (byte)'h', (byte)'i', (byte)'j', (byte)'k', (byte)'l', (byte)'m', (byte)'n',
      (byte)'o', (byte)'p', (byte)'q', (byte)'r', (byte)'s', (byte)'t', (byte)'u', 
      (byte)'v', (byte)'w', (byte)'x', (byte)'y', (byte)'z',
      (byte)'0', (byte)'1', (byte)'2', (byte)'3', (byte)'4', (byte)'5', 
      (byte)'6', (byte)'7', (byte)'8', (byte)'9', (byte)'-', (byte)'_'
    };
	
    // Used in decoding URL and Filename safe dialects of Base64.
    private final static byte[] _URL_SAFE_DECODABET = {
      -9,-9,-9,-9,-9,-9,-9,-9,-9,                 // Decimal  0 -  8
      -5,-5,                                      // Whitespace: Tab and Linefeed
      -9,-9,                                      // Decimal 11 - 12
      -5,                                         // Whitespace: Carriage Return
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 14 - 26
      -9,-9,-9,-9,-9,                             // Decimal 27 - 31
      -5,                                         // Whitespace: Space
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,              // Decimal 33 - 42
      -9,                                         // Plus sign at decimal 43
      -9,                                         // Decimal 44
      62,                                         // Minus sign at decimal 45
      -9,                                         // Decimal 46
      -9,                                         // Slash at decimal 47
      52,53,54,55,56,57,58,59,60,61,              // Numbers zero through nine
      -9,-9,-9,                                   // Decimal 58 - 60
      -1,                                         // Equals sign at decimal 61
      -9,-9,-9,                                   // Decimal 62 - 64
      0,1,2,3,4,5,6,7,8,9,10,11,12,13,            // Letters 'A' through 'N'
      14,15,16,17,18,19,20,21,22,23,24,25,        // Letters 'O' through 'Z'
      -9,-9,-9,-9,                                // Decimal 91 - 94
      63,                                         // Underscore at decimal 95
      -9,                                         // Decimal 96
      26,27,28,29,30,31,32,33,34,35,36,37,38,     // Letters 'a' through 'm'
      39,40,41,42,43,44,45,46,47,48,49,50,51,     // Letters 'n' through 'z'
      -9,-9,-9,-9,-9                              // Decimal 123 - 127
      ,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 128 - 139
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 140 - 152
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 153 - 165
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 166 - 178
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 179 - 191
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 192 - 204
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 205 - 217
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 218 - 230
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 231 - 243
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9         // Decimal 244 - 255 
    };

    
 
    // Private methods
	
    private byte[] getAlphabet() {
        if (urlsafe)
            return _URL_SAFE_ALPHABET;
        else 
        	return _STANDARD_ALPHABET;
    }

    private final byte[] getDecodabet() {
        if (urlsafe)
            return _URL_SAFE_DECODABET;
        else
            return _STANDARD_DECODABET;
    }
    
    /**
     * Encodes up to three bytes of the array source
     * and writes the resulting four Base64 bytes to destination.
    */
    private byte[] encode3to4(byte[] source, int srcOffset, int numSigBytes,
    		byte[] destination, int destOffset)
    {
    	byte[] ALPHABET = getAlphabet(); 
        //           1         2         3  
        // 01234567890123456789012345678901 Bit position
        // --------000000001111111122222222 Array position from threeBytes
        // --------|    ||    ||    ||    | Six bit groups to index ALPHABET
        //          >>18  >>12  >> 6  >> 0  Right shift necessary
        //                0x3f  0x3f  0x3f  Additional AND
        
        // Create buffer with zero-padding if there are only one or two
        // significant bytes passed in the array.
        // We have to shift left 24 in order to flush out the 1's that appear
        // when Java treats a value as negative that is cast from a byte to an int.
        int inBuff =   ( numSigBytes > 0 ? ((source[ srcOffset     ] << 24) >>>  8) : 0 )
                     | ( numSigBytes > 1 ? ((source[ srcOffset + 1 ] << 24) >>> 16) : 0 )
                     | ( numSigBytes > 2 ? ((source[ srcOffset + 2 ] << 24) >>> 24) : 0 );
        switch( numSigBytes )
        {
            case 3:
                destination[ destOffset     ] = ALPHABET[ (inBuff >>> 18)        ];
                destination[ destOffset + 1 ] = ALPHABET[ (inBuff >>> 12) & 0x3f ];
                destination[ destOffset + 2 ] = ALPHABET[ (inBuff >>>  6) & 0x3f ];
                destination[ destOffset + 3 ] = ALPHABET[ (inBuff       ) & 0x3f ];
                return destination;
                
            case 2:
                destination[ destOffset     ] = ALPHABET[ (inBuff >>> 18)        ];
                destination[ destOffset + 1 ] = ALPHABET[ (inBuff >>> 12) & 0x3f ];
                destination[ destOffset + 2 ] = ALPHABET[ (inBuff >>>  6) & 0x3f ];
                destination[ destOffset + 3 ] = EQUALS_SIGN;
                return destination;
                
            case 1:
                destination[ destOffset     ] = ALPHABET[ (inBuff >>> 18)        ];
                destination[ destOffset + 1 ] = ALPHABET[ (inBuff >>> 12) & 0x3f ];
                destination[ destOffset + 2 ] = EQUALS_SIGN;
                destination[ destOffset + 3 ] = EQUALS_SIGN;
                return destination;
                
            default:
                return destination;
        }   // end switch
    }   // end encode3to4

    /**
     * Decodes four bytes from array source
     * and writes the resulting bytes (up to three of them)to destination.
     */
    private int decode4to3(byte[] source, int srcOffset, byte[] destination,
    		int destOffset)
    {        
        // Lots of error checking and exception throwing
        if( source == null ){
            throw new NullPointerException( "Source array was null." );
        }   // end if
        if( destination == null ){
            throw new NullPointerException( "Destination array was null." );
        }   // end if
        if( srcOffset < 0 || srcOffset + 3 >= source.length ){
            throw new IllegalArgumentException( String.format(
            "Source array with length %d cannot have offset of %d and still process four bytes.", source.length, srcOffset ) );
        }   // end if
        if( destOffset < 0 || destOffset +2 >= destination.length ){
            throw new IllegalArgumentException( String.format(
            "Destination array with length %d cannot have offset of %d and still store three bytes.", destination.length, destOffset ) );
        }   // end if
        byte[] DECODABET = getDecodabet(); 
        // Example: Dk==
        if( source[ srcOffset + 2] == EQUALS_SIGN ) {
            int outBuff =   ( ( DECODABET[ source[ srcOffset    ] ] & 0xFF ) << 18 )
                          | ( ( DECODABET[ source[ srcOffset + 1] ] & 0xFF ) << 12 );
            destination[ destOffset ] = (byte)( outBuff >>> 16 );
            return 1;
        }
        // Example: DkL=
        else if( source[ srcOffset + 3 ] == EQUALS_SIGN ) {
            int outBuff =   ( ( DECODABET[ source[ srcOffset     ] ] & 0xFF ) << 18 )
                          | ( ( DECODABET[ source[ srcOffset + 1 ] ] & 0xFF ) << 12 )
                          | ( ( DECODABET[ source[ srcOffset + 2 ] ] & 0xFF ) <<  6 );
            destination[ destOffset     ] = (byte)( outBuff >>> 16 );
            destination[ destOffset + 1 ] = (byte)( outBuff >>>  8 );
            return 2;
        }
        // Example: DkLE
        else {
            int outBuff =   ( ( DECODABET[ source[ srcOffset     ] ] & 0xFF ) << 18 )
                          | ( ( DECODABET[ source[ srcOffset + 1 ] ] & 0xFF ) << 12 )
                          | ( ( DECODABET[ source[ srcOffset + 2 ] ] & 0xFF ) <<  6)
                          | ( ( DECODABET[ source[ srcOffset + 3 ] ] & 0xFF )      );
            destination[ destOffset     ] = (byte)( outBuff >> 16 );
            destination[ destOffset + 1 ] = (byte)( outBuff >>  8 );
            destination[ destOffset + 2 ] = (byte)( outBuff       );

            return 3;
        }
    }
 
    
    // Public methods
    
    /**
     *Decode a byte array of Base64 characters into a byte array. Tab, CR, LF and Space
     *characters in the data are ignored, invalid Base64 characters throw an exception.
     */
    public byte[] DecodeBtoB(byte[] b64bytes, int offset, int length)
    {
        if( length < 4 )
        {
            throw new RuntimeException( 
            "Base64-encoded string must have at least four characters!");
        }        
        byte[] DECODABET = getDecodabet();	
        int    len34   = length * 3 / 4;       // Estimate on array size
        byte[] outBuff = new byte[ len34 ]; // Upper limit on size of output
        int    outBuffPosn = 0;             // Keep track of where we're writing
        byte[] b4        = new byte[4];     // Four byte buffer from source, eliminating white space
        int    b4Posn    = 0;               // Keep track of four byte input buffer
        int    i         = 0;               // Source array counter
        byte   sbiDecode = 0;               // Special value from DECODABET
        
        for( i = offset; i < offset+length; i++ )
        {  // Loop through source            
            sbiDecode = DECODABET[ b64bytes[i]&0xFF ];            
            // White space, Equals sign, or legit Base64 character
            // Note the values such as -5 and -9 in the
            // DECODABETs at the top of the file.
            if( sbiDecode >= WHITE_SPACE_ENC )  {
                if( sbiDecode >= EQUALS_SIGN_ENC ) {
                    b4[ b4Posn++ ] = b64bytes[i];         // Save non-whitespace
                    if( b4Posn > 3 ) {                  // Time to decode?
                        outBuffPosn += decode4to3( b4, 0, outBuff, outBuffPosn);
                        b4Posn = 0;                        
                        // If that was the equals sign, break out of 'for' loop
                        if( b64bytes[i] == EQUALS_SIGN ) {
                            break;
                        }   // end if: equals sign
                    }   // end if: quartet built
                }   // end if: equals sign or better
            }   // end if: white space, equals sign or better
            else {
                // There's a bad input character in the Base64 stream.
                throw new RuntimeException(String.format(
                "Bad Base64 input character decimal %d in array position %d", ((int)b64bytes[i])&0xFF, i ) );
            }   // end else: 
        }   // each input character
                                   
        byte[] out = new byte[ outBuffPosn ];
        System.arraycopy( outBuff, 0, out, 0, outBuffPosn ); 
        return out;
    }
	
    // Public methods
    
    /** 
     *Decode an array of Base64 characters into a string. Tab, CR, LF and Space
     *characters in the data are ignored, invalid Base64 characters throw an exception.
     *Encoding defines the encoding of the byte data of the original coded string.
     */   
    public String DecodeBtoS(byte[] b64bytes,  int offset, int length, String encoding) throws Exception
    {
        return new String(DecodeBtoB( b64bytes, offset, length), encoding);
    }
    
    /** 
     *Decode a String of Base64 characters into a byte array. Tab, CR, LF and Space
     *characters in the string are ignored, invalid Base64 characters throw an exception.
     */
    public byte[] DecodeStoB(String b64string) throws Exception
    {        
        byte[] bytes;
        bytes = b64string.getBytes("US-ASCII");
        bytes = DecodeBtoB( bytes, 0, bytes.length );    
        return bytes;
    }
   
    /** 
     *Decode a string of Base64 characters into a string.  Tab, CR, LF and Space
     *characters in the data are ignored, invalid Base64 characters throw an exception.
     *Encoding defines the encoding of the byte data of the original coded string.
     */
    public String DecodeStoS(String b64string, String encoding) throws Exception
    {        
        byte[] bytes;
        bytes = b64string.getBytes("US-ASCII");
        bytes = DecodeBtoB( bytes, 0, bytes.length);    
        return new String(bytes, encoding);
    }   
    


    /**
     *Encode a byte array into a byte array of Base64 characters.
     */
    public byte[] EncodeBtoB(byte[] data, int offset, int length)
    {
        int encLen = ( length / 3 ) * 4 + ( length % 3 > 0 ? 4 : 0 ); // Bytes needed for actual encoding
        if( breaklines ){
            encLen += encLen / maxlinelength; // Plus extra newline characters
        }
        byte[] outBuff = new byte[ encLen ];
        int d = 0;
        int e = 0;
        int len2 = length - 2;
        int lineLength = 0;
        for( ; d < len2; d+=3, e+=4 ) {
            encode3to4( data, d+offset, 3, outBuff, e);
            lineLength += 4;
            if( breaklines && lineLength >= maxlinelength )
            {
                outBuff[e+4] = NEW_LINE;
                e++;
                lineLength = 0;
            } 
        }
        if( d < length ) {
            encode3to4( data, d+offset, length - d, outBuff, e);
            e += 4;
        }   // end if: some padding needed
        // Only resize array if we didn't guess it right.
        if( e <= outBuff.length - 1 ){
            // If breaking lines and the last byte falls right at
            // the line length (76 bytes per line), there will be
            // one extra byte, and the array will need to be resized.
            // Not too bad of an estimate on array size, I'd say.
            byte[] finalOut = new byte[e];
            System.arraycopy(outBuff,0, finalOut,0,e);
            return finalOut;
        } else {
            return outBuff;
        }
    }
    
    /**
     *Encode a byte array into a string of Base64 characters.
     */
    public String EncodeBtoS(byte[] data, int offset, int length) throws Exception
    {
        byte[] encoded = EncodeBtoB(data, offset, length);
		return new String(encoded, "US-ASCII");
    }
   
 
    /**
     *Encode a string into a byte array of  Base64 characters. The string is converted
     *into an array of bytes according to encoding and then coded as Base64. 
     */    
    public byte[] EncodeStoB(String data, String encoding) throws Exception
    {
    	byte[] encoded;
		encoded = data.getBytes(encoding);
        encoded = EncodeBtoB(encoded, 0, encoded.length);
        return encoded;
    }
    
     /**
     *Encode a string into a string of Base64 characters. The string is converted
     *into an array of bytes according to encoding and then coded as Base64. 
     */    
   public String EncodeStoS(String data, String encoding) throws Exception
    {
		return new String(EncodeStoB(data, encoding), "US-ASCII");
    }
  
    
	// Public properties
   
   
	/**
	 *Gets or sets whether line breaks are added to the Base64 output when encoding.
	 */
	public boolean getBreakLines()
   	{
   		return breaklines;
   	}
  	public void setBreakLines(boolean breaklines)
   	{
   		this.breaklines = breaklines;
   	}
 
	/**
	 *Returns the version number of the library.
	 */
  	public double getVersion()
	{
		return version;
	}
 	
	/**
	 *Gets or sets the number of Base64 characters placed on each line when line breaks
	 *are being added to the Base64 output when encoding.
	 */
	public int getLineLength()
   	{
   		return maxlinelength;
   	}
  	public void setLineLength(int linelength)
   	{
   		maxlinelength = linelength;
   	}
  	
	/**
	 *Gets or sets whether the Base64 encoding and decoding is done in a "URL safe" manner.
	 *In "URL safe mode" the "plus" character in the Base64 becomes a  "hyphen" and  the
	 *"slash" character becomes an "underscore".
	 */
	public boolean getUrlSafe()
   	{
   		return urlsafe;
   	}
  	public void setUrlSafe(boolean urlsafe)
   	{
   		this.urlsafe = urlsafe;
   	}
} 
